import json
import math
from math import isclose
from typing import Tuple

import astropy.units as u
import hypothesis.strategies as st
import numpy as np
import pytest
import requests_mock
from astropy.units.quantity import Quantity
from astropy.utils.data import get_pkg_data_filename
from hypothesis.core import given

from kanon.tables import HTable
from kanon.tables.htable_reader import DISHAS_REQUEST_URL
from kanon.units import Sexagesimal


class TestBasedHTable:

    @requests_mock.Mocker(kw="mock")
    def test_read(self, **kwargs):

        kwargs["mock"].get(requests_mock.ANY, json={"error": "Non existing id"})

        path = get_pkg_data_filename('data/table_content-180.json')
        with open(path, "r") as f:
            content = json.load(f)
        kwargs["mock"].get(DISHAS_REQUEST_URL.format(180), json=content)

        table: HTable = HTable.read(180, format="dishas")

        assert table["Mean Argument of the Sun"].unit is u.degree
        assert table["Mean Argument of the Sun"].basedtype is Sexagesimal
        assert table["Entries"].significant == 2

        assert table.loc[Sexagesimal(1)] == table[0]

        assert table.loc[Sexagesimal(3)]["Entries"].equals(Sexagesimal("-00 ; 06,27"))

        assert table.loc[Sexagesimal(3)]["Entries"].equals(table.get(Sexagesimal(3), with_unit=False))

        assert len(table.symmetry) == 1
        sym = table.symmetry[0]

        assert table.get(1) == -table.get(37)

        assert sym.symtype == "mirror"

        assert "Sexagesimal" in repr(table)

        with pytest.raises(FileNotFoundError):
            HTable.read(181, format="dishas")

    gen_table_strategy = st.builds(
        HTable,
        st.lists(
            st.tuples(
                st.from_type(Sexagesimal).map(lambda x: x.resize(2).truncate()),
                st.from_type(Sexagesimal).map(lambda x: x.resize(2).truncate())
            ), min_size=2, max_size=7, unique_by=(lambda x: x[0])
        ).map(lambda x: list(zip(*sorted(x)))),
        names=st.just(("A", "B")),
        index=st.just("A"),
        dtype=st.just([object, object])
    )

    @given(gen_table_strategy.flatmap(
        lambda x: st.tuples(st.floats(
            min_value=float(min(x["A"])),
            max_value=float(max(x["A"])),
            allow_nan=False
        ), st.just(x))))
    def test_interpolation(self, hypo: Tuple[float, HTable]):
        key, tab = hypo
        fres = np.interp(key, [float(x) for x in tab.columns[0]], [float(x) for x in tab.columns[1]])
        sres = tab.get(key)
        assert isclose(fres, float(sres), abs_tol=1e9)

    @given(gen_table_strategy)
    def test_quantity(self, tab: HTable):
        tab["A"].unit = u.degree
        tab["B"].unit = u.degree

        value = tab.get(tab["A"][0])

        assert isinstance(value, Quantity)
        assert value.unit is u.degree
        assert isinstance(value.value, Sexagesimal)

    @given(gen_table_strategy)
    def test_loc_slice(self, tab: HTable):
        sliced = tab.loc[tab["A"][1]:]
        assert not isinstance(sliced, HTable) or len(sliced) == len(tab) - 1

    def test_based_fill(self):
        sin_table = HTable([
            list(Sexagesimal.range(91)),
            [round(Sexagesimal.from_float(math.sin(x * math.pi / 180), 3)) for x in range(91)]
        ], names=("Arg", "Val"), index="Arg")
        sin_table_grid = sin_table[
            [i for i in range(91) if i <= 60 and i % 12 == 0 or i > 60 and i % 6 == 0]
        ]
        sin_table_pop_euclidean = sin_table_grid.populate(list(Sexagesimal.range(91)))
        sin_table_pop_euclidean.fill("distributed_convex")



from typing import Callable, Dict, List, Tuple, cast

import astropy.units as u
import requests
from astropy.units.core import Unit

from kanon.tables.symmetries import Symmetry
from kanon.units import Historical, Sexagesimal
from kanon.units.radices import BasedReal
from kanon.utils.types.dishas import (DSymmetry, NumberType, TableContent,
                                      UnitType)
from kanon.utils.types.number_types import Real

from .htable import HTable

_dishas_fields = '","'.join([
    "source_value_original",
    "argument1_name",
    "argument1_number_unit",
    "argument1_significant_fractional_place",
    "argument1_type_of_number",
    "entry_significant_fractional_place",
    "entry_number_unit",
    "entry_type_of_number",
    "symmetries",
    "edited_text"
])
DISHAS_REQUEST_URL = f'https://dishas.obspm.fr/elastic-query\
?index=table_content&hits=true&id={{}}&source=["{_dishas_fields}"]'


def read_sexag_array(array: List[str], shift: int) -> BasedReal:
    sign = -1 if array[0][0] == "-" else 1
    return Sexagesimal(*(abs(int(v)) for v in array), sign=sign) >> shift


def read_intsexag_array(array: List[str], _: int) -> BasedReal:
    integer = int(array[0])
    if len(array) == 1:
        return Sexagesimal.from_int(integer)
    return integer + (read_sexag_array(array[1:], 0) >> len(array) - 1)


def read_historical(array: List[str], _: int) -> BasedReal:
    integer = int(array[0])
    if len(array) == 2:
        return integer * 30 + Historical(array[1])
    return Historical(*(int(x) for x in array))


def read_table_dishas(requested_id: str) -> HTable:

    res: TableContent = requests.get(
        DISHAS_REQUEST_URL.format(int(requested_id)),
    ).json()
    if not res or "error" in res:
        raise FileNotFoundError(
            f'{requested_id} ID not found in DISHAS database')

    values = res["source_value_original"]

    number_reader: Dict[NumberType, Callable[[List[str], int], Real]] = {
        "sexagesimal": read_sexag_array,
        "floating sexagesimal": read_sexag_array,
        "integer and sexagesimal": read_intsexag_array,
        "historical": read_historical
    }

    unit_reader: Dict[UnitType, Unit] = {
        "degree": u.degree,
        "day": u.day
    }

    arg_unit = res["argument1_number_unit"]
    arg_shift = int(res["argument1_significant_fractional_place"])
    arg_reader = number_reader.get(res["argument1_type_of_number"], lambda x, _: x)

    entry_unit = res["entry_number_unit"]
    entry_shift = int(res["entry_significant_fractional_place"])
    entry_reader = number_reader.get(res["entry_type_of_number"], lambda x, _: x)

    args = [arg_reader(v["value"], arg_shift) for v in values["args"]["argument1"]]
    entries = [entry_reader(v["value"], entry_shift) for v in values["entry"]]

    symmetry_raw = res["symmetries"]

    def build_sym(data: DSymmetry) -> Symmetry:
        symmetry = Symmetry(data["symtype"], data["offset"], sign=data["sign"])

        if source_data := data["source"]:
            source = tuple(arg_reader(v, arg_shift) for v in source_data)
            assert len(source) == 2
            symmetry.source = cast(Tuple[Real, Real], source)

        if target_data := data["target"]:
            symmetry.targets = [arg_reader(v, arg_shift) for v in target_data]

        return symmetry

    symmetries = [build_sym(sym) for sym in symmetry_raw]

    table = HTable(
        [args, entries],
        names=(res["argument1_name"], "Entries"),
        index=(res["argument1_name"]),
        units=[unit_reader.get(arg_unit), unit_reader.get(entry_unit)],
        dtype=[object, object],
        symmetry=symmetries,
        meta=res["edited_text"]
    )

    return table



from typing import Any, Dict

import pytest
from astropy.units import degree
from papermill.execute import execute_notebook

from kanon.units import Sexagesimal


class TestNotebooks:

    def get_nb(self, name: str, params: Dict[str, Any]):
        nb = execute_notebook(f'./examples/{name}.ipynb', "-", parameters=params)
        return nb

    @pytest.mark.parametrize("params,result", [
        ({"year": 1327, "month": 7, "day": 3}, "1,47;18,49"),
        ({"year": 1691, "month": 9, "day": 9}, "2,55;31,33"),
        ({"year": 998, "month": 10, "day": 2}, "03,13;25,33"),
        ({"year": 1998, "month": 2, "day": 10}, "05,34;29,06"),
    ])
    def test_sun_true_position(self, params, result):
        data = self.get_nb("sun_true_position", params).cells[-1].outputs[0].data["text/latex"]
        assert data == (Sexagesimal(result) * degree)._repr_latex_()

    @pytest.mark.parametrize("params,result", [
        ({"OBLIQUITY": "23;51,20"}, ("02;01,12,38", "23;51,20,22")),
        ({"OBLIQUITY": "0"}, ("0", "0")),
    ])
    def test_declination(self, params, result):
        data: str = self.get_nb("declination", params).cells[7].outputs[0].data["text/html"]
        lines = data.split("<tr>")
        line5 = [li for li in lines if "<td>05 ;" in li][0]
        line90 = [li for li in lines if "<td>01,30 ;" in li][0]
        assert repr(Sexagesimal(result[0])).strip() in line5
        assert repr(Sexagesimal(result[1])).strip() in line90

    @pytest.mark.parametrize("params,result", [
        ({"OBLIQUITY": "23;51,20"}, ("0.3111", "-3.0222")),
        ({"OBLIQUITY": "0"}, ("54610.2778", "55083.2889")),
    ])
    def test_ptolemy_viz(self, params, result):
        output = "".join(o["text"] for o in self.get_nb("ptolemy_viz", params).cells[9].outputs)
        assert f"mean : {result[0]}" in output
        assert f"mean : {result[1]}" in output



import math

import hypothesis.strategies as st
import pytest
from hypothesis import given

from kanon.calendars import Calendar, Date
from kanon.calendars.calendars import Era, Julian, float_to_hm, hm_to_float


class TestCalendars:

    def test_init(self):
        era = Era("test", 10)

        assert era.days_from_epoch(20) == 10

        new_cal = Julian(era)
        assert new_cal.era == era
        assert new_cal.name in Calendar.registry

        assert new_cal.__repr__() == "Calendar(Julian test)"

        with pytest.raises(ValueError):
            Julian(era)

        with pytest.raises(ValueError):
            new_cal.jdn_at_ymd(1, 50, 1)
        with pytest.raises(ValueError):
            new_cal.jdn_at_ymd(1, 1, 50)
        with pytest.raises(ValueError):
            new_cal.jdn_at_ymd(0, 1, 10)

    def test_first_day(self):
        for cal in Calendar.registry.values():
            assert cal.jdn_at_ymd(1, 1, 1) == cal.era.epoch
            assert Date(cal, (1, 1, 1)).days_from_epoch() == 0

    def test_from_julian(self):
        cal = Calendar.registry["Julian A.D."]
        days = cal.jdn_at_ymd(2, 1, 1)

        assert cal.from_julian_days(days).jdn == days
        assert cal.from_julian_days(640234).jdn == 640234
        assert cal.from_julian_days(cal.era.epoch - 1).jdn == cal.era.epoch - 1

    @given(st.integers(1, 10000), st.integers(1, 12), st.integers(1, 28))
    def test_dates_hypothesis(self, y: int, m: int, d: int):
        cal = Calendar.registry["Julian A.D."]

        date = Date(cal, (y, m, d))

        assert cal.from_julian_days(date.jdn) == date
        assert cal.get_time(y, m, d) == date.to_time()

    def test_conversion(self):
        arabic = Calendar.registry["Arabic Civil Hijra"]
        julian = Calendar.registry["Julian A.D."]

        date = Date(arabic, (1, 1, 1))
        conversion = date.to_calendar(julian)
        assert conversion.ymd == (622, 7, 16)

    def test_date(self):
        cal = Calendar.registry["Julian A.D."]
        date = Date(cal, (-4713, 1, 1))

        assert date.to_time().to_value("jd") == 0
        assert str(date) == "1 Ianuarius -4713 A.D. in Julian 12:00"

        date = Date(cal, (-1, 1, 1))

        assert date.to_time().to_value("jd") == 1721058
        assert str(date) == "1 Ianuarius -1 A.D. in Julian 12:00"

        date = Date(cal, (20, 3, 12))
        assert (date + 1).jdn == date.jdn + 1
        assert (date - 1).jdn == date.jdn - 1

        assert "13:30" in str(Date(cal, (1, 1, 1), hm_to_float(13, 30)))

        with pytest.raises(ValueError) as err:
            Date(cal, (1, 1, 1), -1)

        assert "Time must be" in str(err)

    def test_jdn_at_ymd(self):
        cal = Calendar.registry["Julian A.D."]

        assert cal.jdn_at_ymd(-1, 12, 31) + 1 == cal.jdn_at_ymd(1, 1, 1)
        assert cal.jdn_at_ymd(-2, 12, 31) + 1 == cal.jdn_at_ymd(-1, 1, 1)

        assert cal.jdn_at_ymd(61, 1, 1) == 1743339

        assert cal.jdn_at_ymd(60, 12, 31) != cal.jdn_at_ymd(61, 1, 1)

    def test_persian(self):
        cal_normal = Calendar.registry["Persian Yazdigird Andarjah at the end"]
        cal_variant = Calendar.registry["Persian Yazdigird Andarjah after Ābān"]

        assert len(cal_normal.months) == len(cal_variant.months)
        assert cal_normal.months[-1].name == cal_variant.months[8].name

    def test_dates_fraction(self):
        cal = Calendar.registry["Julian A.D."]

        assert cal.from_julian_days(5000.6) != cal.from_julian_days(5000)

        date = Date(cal, (5, 2, 3), 0.4)

        assert math.isclose((date + 1).frac, 0.4)

        assert (date + 1).ymd == (5, 2, 4)

    def test_time_convert(self):
        assert float_to_hm(hm_to_float(12, 30)) == (12, 30)

        with pytest.raises(ValueError) as err:
            hm_to_float(-1, 5)

        assert "Incorrect" in str(err)

        with pytest.raises(ValueError) as err:
            float_to_hm(2)

        assert "Incorrect" in str(err)
